from geosquizzy.structure import GeoJSON
from tests.getdata import get_geojson


if __name__ == "__main__":
    geo = GeoJSON(geojson_doc_type="FeatureCollection")
    """
    artificially generated geojson data
    """
    # data = get_geojson(url="https://raw.githubusercontent.com/LowerSilesians/geo-squizzy/master/build_big_data/test_data/dump1000.json")
    """
    live geojson data
    """
    # data = get_geojson(url="https://raw.githubusercontent.com/LowerSilesians/geo-squizzy/master/build_big_data/test_data/ExampleDataPoint.json")
    data = get_geojson(path="/home/ing/PycharmProjects/geo-squizzy/geosquizzy/build_big_data/data/dump5000.json")
    geo.start(geojson=data, is_doc=True)

    for x in geo.get_results():
        print(x, '\n')


import re
import time
"""
@TOKENS -> dict() which contains all values that can occur during tokenizing process
@GRAMMAR_RULES -> dict() which contains some grammars(some specific relationships between tokens which
                         always occur, there are not individual cases)
"""

TOKENS = {
    'VALUE': r'[\w,\d]',
    'COMMA': r'[,]',
    'COLON': r'[:]',
    'QUOTATION_MARKS': r'["]',
    'WHITE_SPACE': r'[\s]',
    'SQUARE_BRACKET_O': r'[\[]',
    'SQUARE_BRACKET_E': r'[]]',
    'CURLY_BRACKET_O': r'[\{]',
    'CURLY_BRACKET_E': r'[}]',
}

SEQUENCES = {
    'KEY_O': r'[,]+[\{]+["]+|[\{]+["]+',
    'KEY_E': r'[:]+[\s]*["]+'
}


class JsonTokenizer:
    """
    usage -> for now we assume that a string represents full geojson
             features array in string format
    name  -> JsonTokenizer
    """

    def __init__(self, *args, **kwargs):
        """
        kwargs['structure'] -> instance of class FeaturesTree
                self.structure and self.tree from class GeoJSON now have the same
                reference point

        self.STATUS ->Integer() would mean which grammar rule was applied and
                                when status is 0 could mean that there is no
                                grammar rule currently applied
        """
        self.structure = kwargs['structure']
        self.tokens = TOKENS
        self.sequences = SEQUENCES
        self.STATUS = 0
        self.LEVEL = 0

    # GRAMMAR_RULES #
    def root_rule(self):
        """
        if first char is '['
        we open feature json array

        we are omitting meet char no status is set
        :return:
        """
        pass

    def nested_rule(self):
        self.STATUS = 2

    def key_rule(self):
        """
        if combination of chars ,{", {", was meet
        """
        self.STATUS = 3

    def value_rule(self):
        """
        if combination of chars :(?:[\s]*)", was meet
        it will end when char " is presented
        """
        self.STATUS = 4

    def catch_key(self, data=None):
        word = ''
        for i, x in enumerate(data):
            #print(i, x)
            #assert False
            if not re.match(self.tokens['QUOTATION_MARKS'], x):
                word += x
            else:
                data = data[i + 1:-1]
                break
        return word, data

    def run_tokenizer(self, data=None):
        start = time.time()
        words_list = []
        word = str()
        for i, x in enumerate(data):
            # if self.STATUS == 1 and not re.match(self.tokens['QUOTATION_MARKS'], x):
            #     word += x
            # if re.match(self.tokens['QUOTATION_MARKS'], x):
            #     if self.STATUS == 0:
            #         self.STATUS = 1
            #     elif self.STATUS == 1:
            #         #print(word, '\n')
            #         word = str()
            #         self.STATUS = 0
            if self.STATUS == 0 and re.match(
                    self.tokens['CURLY_BRACKET_O'], x):
                """new json object"""
                self.STATUS = 1
            elif self.STATUS == 1 and re.match(self.tokens['QUOTATION_MARKS'], x):
                """json object key starting"""
                print(data.__len__())
                word, data = self.catch_key(data[i + 1:-1])
                print(word, data.__len__(), data[0])
                # next option is
                # 1" VALUE "
                # 2{ nested value and we could go back to second if
                # 3[ which could have few possibilities
                # 1 [1123,12323] -> coordinates
                # 2 {} collection of objects
                # 3 [[],[]] collection of collectiond
                #assert False
                pass
        end = time.time()
        print(data.__len__())
        print(end - start)



try:
    from urllib.request import urlopen
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
    from urllib import urlopen


def get_geojson(url=None, path=None):
    """
    get url data and decode it
    """
    if url is not None:
        res = urlopen(url=url)
        return res.read().decode('utf-8')
        # http://stackoverflow.com/questions/30627937/tracebaclk-attributeerroraddinfourl-instance-has-no-attribute-exit
    else:
        file = open(path, 'r')
        return file.read()




from utils import get_random_number, get_random_string


class CITY:
    def __init__(self):
        self.data = dict({"name": str(),
                          "code": int(),
                          "country": str(),
                          "citizens": int()})

    def get_random_data(self):
        self.data.clear()
        self.data['code'] = get_random_number(0, 10000)
        self.data['citizens'] = get_random_number(500, 500000)
        self.data['name'] = get_random_string(size=8)
        self.data['country'] = get_random_string(size=7)
        return self.data


import re


def is_float(arg):
    return isinstance(arg, float)


def is_geojson_doc(arg):
    pattern = r'features'
    return bool(re.match(pattern, arg))


def get_string_slice(patterns, arg):
    """
    :param patterns: [regex,...]
    :param arg:  string
    :return: wanted string slice
    """
    while True:
        try:
            pattern = patterns.pop()
            parts = re.split(pattern, arg, maxsplit=1)
            arg = (parts[0].__len__() > parts[
                   1].__len__()) and parts[0] or parts[1]
        except IndexError:
            break
    return arg


# TODO some util to remove white spaces around value
# TODO could be use after traversing doc as cleaning function or sth



import random
import json
from models import CITY


class Duplicates:
    def __init__(self):
        self.storage = dict()
        pass


class Feature:
    def __init__(self, *args, **kwargs):
        self.data = dict({"type": "Feature", "properties": dict(), "geometry": {"type": kwargs['type'], "coordinates": []}})
        self.data['properties'] = kwargs['model']().get_random_data()

    def add_coordinates(self, coordinates=None):
        self.data['geometry']['coordinates'] = coordinates


class DataStructure:
    def __init__(self, *args, **kwargs):
        self.data = dict({'type': kwargs['type'], 'features': []})
        self.duplicates = Duplicates()
        self._range = kwargs['coordinates_range']
        self.feature_model = kwargs['feature_model']
        self.feature_type = kwargs['feature_type']
        self.__run__(number=kwargs['features_number'])
        pass

    def __run__(self, number=None):
        self.data['features'] = [self.feature() for x in range(0, number, 1)]
        pass

    def coordinates(self):
        x = random.uniform(self._range[0], self._range[1])
        case = self.duplicates.storage.get(x, None)
        while case is not None:
           x = random.uniform(self._range[0], self._range[1])
           case = self.duplicates.storage.get(x, None)

        self.duplicates.storage[x] = x
        return x

    def feature(self):
        feature = Feature(type=self.feature_type, model=self.feature_model)
        feature.add_coordinates(coordinates=[self.coordinates(), self.coordinates()])
        return feature.data


if __name__ == "__main__":
    geo = DataStructure(type="FeatureCollection",
                        feature_type="Point",
                        coordinates_range=[float(-200), float(200)],
                        features_number=500000,
                        feature_model=CITY)
    geo_json = json.dumps(geo.data)
    f = open("data/dump500000.json", "w")
    f.write(geo_json)
    f.close()


class FeatureParser:
    """
    Single Unit which is a part of geojson features array
    """

    def __init__(self):
        pass


class FeaturesParser:
    """
    GeoJSON features parser representation
    """

    def __init__(self):
        pass



MESSAGES = dict({
    "1": "\n\nData provided for method __init__() of class Feature \n"
         "was not correct to create dict() object",
    "2": "\n\nFeature object is not a valid geojson feature object, \n"
         "one of the following failed, geometry, properties or type field are missing \n",
    "3": "\n\nCoordinates Array Items must be of float type\n"
})


import re
import copy


class GeojsonFiniteStateMachine:
    # TODO create separation between fsm core logic and temp data
    # usage/functionality

    def __init__(self, *args, **kwargs):
        """
        @kwargs['structure'] Nodes/Tree json keys model representation
        """
        self.structure = kwargs['structure']
        self.MBC = [0, 0, 0, 0]
        self.stack_structure = []
        self.command = [None, '0', [0, 0, 0, 0], -1]
        self.words = []
        self.values = []
        self.key = ''
        self.lib_commands = {
            '0{': ['EXP', '01', [0, 1, 0, 0], 0],
            '0:': ['VAL', '0000', [1, 0, 0, 0], 0],
            '0,': ['DEL', '01', [0, 0, 0, 1], 0],
            '0}': ['DEL', '11', [0, 0, 0, 1], 0],
            '0]': ['DEL', '11', [0, 0, 0, 1], 0],
            '0VAL"': ['EXP', '10', [0, 1, 0, 0], 0],
            '0VALDIG': ['DIG', '01', [0, 1, 0, 0], 0],
            '0VAL{': ['EXP', '01', [0, 1, 0, 0], 0],
            '0VAL[': ['ARR', '10', [1, 0, 0, 0], 0],
            '0ARR"': ['EXP', '11', [0, 1, 0, 0], 0],
            '0ARR{': ['EXP', '01', [0, 1, 0, 0], 0],
            '0ARRDIG': ['DIG', '11', [0, 1, 0, 0], 0]
        }

    def __create_unique_id__(self, key):
        """
        Creating unique ids which represent keys in structure nodes,
        allowing us to store the same key names but which exist
        on different levels of doc structure
        """
        if key == 0:
            'parent'
            try:
                name = self.words[-2]
                return name + str(self.words.__len__() - 1)
            except IndexError:
                return None
        elif key == 1:
            'children'
            return self.words[-1] + str(self.words.__len__())

    def __get__command(self, char=None, mbc=None):
        """
        @:return one of command from lib_commands
        for security we added deepcopy, because somewhere in the code,
        the lib_commands because of reference was changed,
        """
        # TODO self.lib_commands should be immutable object, and all mutables
        # should be placed in different structure
        try:
            return copy.deepcopy(self.lib_commands[mbc + char])
        except KeyError:
            return self.command

    def __modify_structure_stack__(self, **kwargs):
        if kwargs['arg'] == '[':
            self.stack_structure.append('ARR')
        elif kwargs['arg'] == '{':
            self.stack_structure.append('OBJ')
        elif kwargs['arg'] == ']' or kwargs['arg'] == '}':
            try:
                self.stack_structure.pop()
            except IndexError:
                pass

    def __bring_fsm_initial_state(self):
        return [None, '0', [1, 0, 0, 0], -1]

    def __extend_key__(self, **kwargs):
        if kwargs['char'] != '\n':
            self.key += kwargs['char']

    def __add__word__(self):
        self.words.append(re.sub('"', '', self.key))

    def __remove__word__(self):
        try:
            self.words.pop()
        except IndexError:
            pass

    def __add_value__(self):
        self.values.append(re.sub('"', '', self.key))

    def __clean__values__(self):
        self.values = []

    def __save_value__(self):
        self.structure.add_leaf_values(
            id=self.__create_unique_id__(1),
            values=self.values)
        self.key = ''

    def __save__word__(self):
        node = self.structure.prepare_new_leaf(id=self.__create_unique_id__(1),
                                               name=self.words[-1],
                                               level=self.words.__len__(),
                                               parent=self.__create_unique_id__(0))
        self.structure.add_leaf(leaf=node)
        self.key = ''

    def __interpret__(self, **kwargs):
        """
        :param kwargs: arg = character
        :return:None
        """
        if not (self.command[0] is None):
            if self.command[0] == 'VAL':
                # TODO TEMP DIGIT INTERPRET
                if kwargs['char'] == '-' or kwargs['char'].isdigit():
                    self.command = self.__get__command(
                        char=self.command[0] + 'DIG', mbc='0')
                    self.MBC = self.command[2]
                else:
                    self.command = self.__get__command(
                        char=self.command[0] + kwargs['char'], mbc='0')
                    self.MBC = self.command[2]
            elif self.command[0] == 'ARR':
                # TODO TEMP DIGIT INTERPRET
                if kwargs['char'] == '-' or kwargs['char'].isdigit():
                    self.command = self.__get__command(
                        char=self.command[0] + 'DIG', mbc='0')
                    self.MBC = self.command[2]
                else:
                    self.command = self.__get__command(
                        char=self.command[0] + kwargs['char'], mbc='0')
                    self.MBC = self.command[2]
        else:
            self.command = self.__get__command(char=kwargs['char'], mbc='0')
            self.MBC = self.command[2]

    def __write__(self, **kwargs):
        """
        :param kwargs: arg = character
        :return:None
        """
        if self.command[0] == 'EXP':
            if kwargs['char'] == '"':
                # open expression
                self.command[3] += 1
            if self.command[3] > 0:
                # saving expression
                # self.key += kwargs['char']
                self.__extend_key__(**kwargs)
            if self.command[3] == 2:
                # stop saving expression
                self.MBC = [0, 0, 1, 0]
        elif self.command[0] == 'DIG':
            if self.command[1] == '11':
                if kwargs['char'] == ',':
                    self.__add_value__()
                    self.key = ''
                elif kwargs['char'] == ']':
                    self.__add_value__()
                    self.key = ''
                    self.MBC = [0, 0, 1, 0]
                else:
                    self.__extend_key__(**kwargs)
                    # self.key += kwargs['char']
            elif self.command[1] == '01':
                if kwargs['char'] == ',':
                    self.__add_value__()
                    self.key = ''
                    self.MBC = [0, 0, 1, 0]
                elif kwargs['char'] == '}':
                    self.__add_value__()
                    self.key = ''
                    self.MBC = [0, 0, 1, 0]
                else:
                    self.__extend_key__(**kwargs)
                    # self.key += kwargs['char']

    def __save__(self, **kwargs):
        """
        :param kwargs: arg = character
        :return:None
        """
        if self.command[0] == 'EXP':
                # save values " some expression "
            if self.command[1] == '01':
                # add key
                self.__add__word__()
                # save key
                self.__save__word__()
                self.command = self.__bring_fsm_initial_state()
                self.MBC = self.command[2]
            elif self.command[1] == '10':
                self.__add_value__()
                self.__save_value__()
                self.command = self.__bring_fsm_initial_state()
                self.MBC = self.command[2]

        elif self.command[0] == 'DIG':
            # save digits
            if self.command[1] == '11':
                self.__save_value__()
                # self.command[0], self.command[2], self.MBC = None, [1, 0, 0, 0], [1, 0, 0, 0]
                self.command = self.__bring_fsm_initial_state()
                self.MBC = self.command[2]
            elif self.command[1] == '01':
                self.__save_value__()
                if kwargs['char'] == ',':
                    self.command, self.MBC = [
                        'DEL', '01', [
                            0, 0, 0, 1], 0], [
                        0, 0, 0, 1]
                elif kwargs['char'] == '}':
                    self.command, self.MBC = [
                        'DEL', '11', [
                            0, 0, 0, 1], 0], [
                        0, 0, 0, 1]

    def __remove__(self, **kwargs):
        """
        :param kwargs: arg = character
        :return:None
        """
        if self.command[0] == 'DEL':

            if self.command[1] == '01':
                self.__remove__word__()
                self.__clean__values__()
                # switch to write mode
                self.command = ['EXP', '01', [0, 1, 0, 0], 0]
                self.MBC = [0, 1, 0, 0]
            elif self.command[1] == '11':
                # will remove two words, because we are closing object
                try:
                    if self.stack_structure[-1] != 'ARR':
                        self.__remove__word__()
                        self.__clean__values__()
                except IndexError:
                    self.__remove__word__()
                    self.__clean__values__()
                self.command = self.__bring_fsm_initial_state()
                self.MBC = self.command[2]

    def run(self, **kwargs):
        """
        @kwargs['data'] features array
        @self.MBC[0] is set to 1 to run Interpret state as initial one
        """
        self.data = kwargs['data']
        self.MBC[0] = 1

        for i, k in enumerate(self.data):
            # print(self.command, ' ', k, '  ', i)
            # print(self.words)
            # print('\n')
            self.__modify_structure_stack__(arg=k)

            if self.MBC[0] == 1:
                # interpret #
                self.__interpret__(char=k)
            if self.MBC[1] == 1:
                # write #
                self.__write__(char=k)
            if self.MBC[2] == 1:
                # save #
                self.__save__(char=k)
            if self.MBC[3] == 1:
                # delete #
                self.__remove__()

            if i == 330:
                # pass
                # print(self.structure.nodes)
                # TODO 282 should remove value
                # TODO 296 ? not removed words ?
                # TODO 322-323 key not added ?
                # assert False
                pass



from exceptions.messages import MESSAGES


class FeatureSyntaxError(SyntaxError):
    def __init__(self):
        super(FeatureSyntaxError, self).__init__(MESSAGES['1'])


class FeatureStructureError(Exception):
    def __init__(self):
        super(FeatureStructureError, self).__init__(MESSAGES['2'])


class FeatureCoordinatesError(Exception):
    def __init__(self):
        super(FeatureCoordinatesError, self).__init__(MESSAGES['3'])


from setuptools import setup, find_packages
import os

# TODO there is some problem with adding README.md file which have markup tokens
long_description = 'None'
if os.path.exists('README.txt'):
    long_description = open('README.txt').read()

# dev version syntax 0.1.dev4
setup(
    name='geo-squizzy',
    version='0.1.2',
    packages=find_packages(exclude=['research*', 'tests*', 'geosquizzy/testing.py']),
    license='MIT',
    description='GeoJSON-unknown-documents-model-creation',
    long_description=long_description,
    url='https://github.com/LowerSilesians/geo-squizzy',
    author='Igor Miazek, Jakub Miazek',
    author_email='t-32@o2.pl, the@grillazz.com'
)


import random
import string

STRING_CHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'


def get_random_number(arg1, arg2):
    return int(random.uniform(arg1, arg2))


def get_random_string(size=0):
    return ''.join(random.choice(STRING_CHARS) for _ in range(size))


from copy import deepcopy
import geosquizzy.utils as utils
import geosquizzy.tokenizer as tokenizer
import geosquizzy.fsm as fsm


class Tree:

    def __init__(self, *args, **kwargs):
        self.leaf = dict({'id': None, 'name': None, 'children': [], 'level': 0,
                          'parent': None, 'completed': False, 'values': []})
        """
        @self.leaf completed property will announce that the leaf is ready
        """
        self.tree = dict()
        self.nodes = dict()

    def add_leaf(self, leaf=None):
        if leaf['parent'] is None:
            """root"""
            self.nodes[leaf['id']] = leaf
        elif self.nodes.get(leaf['id'], None) is None:
            """new leaf"""
            self.nodes[leaf['id']] = leaf

            if leaf['id'] not in self.nodes[leaf['parent']]['children']:
                """new children"""
                self.nodes[leaf['parent']]['children'].append(leaf['id'])

    def add_leaf_values(self, id=None, values=None):
        self.nodes[id]['values'] = values

    def get_all_leafs_paths(self):
        """we search through all nodes dict structure and for each element
           which have an empty children array we will start to construct search path,
           @query elements
            in order to store more then one key of the same name but which occurred on different nestedness level
            we have to provide unique name for it which is an @id, and for real key search representation we use
            @name attribute
            @id has to be used as well as linking key to describe inheritance
            child @id @name parent=@id | parent @id(child parent)
        """
        paths = []
        for x in self.nodes:
            new_key = {'keys': [], 'values': []}
            if self.nodes[x]['children'].__len__() == 0:
                # print(self.nodes[x]['name'], self.nodes[x]['values'], '\n')
                new_key['keys'].append(self.nodes[x]['name'])
                new_key['values'] = self.nodes[x]['values']
                last_parent = self.nodes[x]['parent']
                # if last parent is None at very begining it mean that it is very low level prop but if doesn't have children
                # then it has to be added as well
                if last_parent is None:
                    paths.append(new_key)
                while not (last_parent is None):
                    new_key['keys'].append(self.nodes[last_parent]['name'])
                    if not (self.nodes[last_parent]['parent'] is None):
                        last_parent = self.nodes[last_parent]['parent']
                    else:
                        paths.append(new_key)
                        last_parent = None
        return paths

    def prepare_new_leaf(self, **kwargs):
        new_leaf = deepcopy(self.leaf)
        new_leaf['id'] = kwargs.get('id', None)
        new_leaf['name'] = kwargs.get('name', None)
        new_leaf['level'] = kwargs.get('level', None)
        new_leaf['parent'] = kwargs.get('parent', None)
        return new_leaf


class FeaturesTree(Tree):

    def __init__(self, *args, **kwargs):
        # super(FeaturesTree, self).__init__(*args, **kwargs)
        Tree.__init__(self, *args, **kwargs)


class GeoJSON:

    def __init__(self, *args, **kwargs):
        self.type = kwargs['geojson_doc_type']
        self.data = dict({"type": self.type, "features": []})
        self.tree = FeaturesTree()
        self.tokenizer = tokenizer.JsonTokenizer(structure=self.tree)
        self.fsm = fsm.GeojsonFiniteStateMachine(structure=self.tree)
        self.geojson = None
        self.percentage = None
        self.is_doc = False

    def start(self, **kwargs):
        """
        kwargs['is_doc'] is optional flag but provided will grow
        the performance, we will not have to check if doc is a chunk or full doc
        """
        self.geojson = kwargs.get('geojson', None)
        self.percentage = kwargs.get('percentage', None)
        self.is_doc = kwargs.get('is_doc', False)
        self.__read_geojson__()

    def get_results(self):
        return self.tree.get_all_leafs_paths()

    def __read_geojson__(self):
        """
        @self.percentage How much of the provided geojson should be checked, default is set to all
        @self.geojson String which can represent both or full geojson document or what will be in favour for future
         improvements it could be a chunk of String which consist a full geojson document

        as main activity it should all the time update self.tree nodes
        """

        if self.is_doc or utils.is_geojson_doc(self.geojson):
            """
            geojson doc mode
            when we are getting a full geojson object we need only to get a features array
            to achieve it we use to regex patterns, which are very general and should give us the [features inside]
            """
            # patterns = [r'(?:[,\s]*)"features":(?:[\s]*)',
            #             r'(?:[,\s]*)"type":(?:[\s]*)"FeatureCollection"(?:[,\s]*)']
            # features_string = utils.get_string_slice(patterns, self.geojson)
            self.fsm.run(data=self.geojson)
        else:
            """
            geojson chunk mode
            """
            pass



from exceptions.exceptions import *
from utils import *


class Feature:

    def __init__(self, *args, **kwargs):
        """
        :param args:
        :param kwargs: data -> appropriate data representation to create dict
        :return: Feature object

        Base Class which represent Feature object with basic data validation
        """
        try:
            self.data = dict(kwargs['data'])
        except Exception:
            """
            Exception will be raised if dict() initialization will fail
            """
            raise FeatureSyntaxError()

        geometry = self.data.get("geometry", None)
        properties = self.data.get("geometry", None)
        feature = self.data.get("type", None)
        if not geometry or not properties or feature is not "Feature":
            """
            Exception will be raised feature object is not an valid feature
            object according to geojson docs
            """
            raise FeatureStructureError()

    def validate(self):
        pass

    def is_feature_type(self):
        return self.data["type"] == self.name


class PointFeature(Feature):

    def __init__(self, *args, **kwargs):
        super(PointFeature, self).__init__(*args, **kwargs)
        self.name = "Point"

    def validate_point(self):
        for x in self.data["geometry"]["coordinates"]:
            if not is_float(x):
                raise FeatureCoordinatesError()
        if not self.is_feature_type():
            raise Exception("\nWrong type provided for PointFeature Class.")

    def validate(self):
        super(PointFeature, self).validate()
        self.validate_point()


class MultiPointFeature(Feature):

    def validate(self):
        pass


class LineStringFeature(Feature):

    def validate(self):
        pass


class MultiLineStringFeature(Feature):

    def validate(self):
        pass


class PolygonFeature(Feature):

    def validate(self):
        pass


class MultiPolygonFeature(Feature):

    def validate(self):
        pass


if __name__ == "__main__":
    data_ok = {"geometry": {"type": "Poin",
                            "coordinates": [-122.93770201248995,
                                            146.32791746493376]},
               "properties": {"code": 4402,
                              "name": "BZgtQyEu",
                              "citizens": 351641,
                              "country": "WKyCMBr"},
               "type": "Feature"}

    test = PointFeature(data=data_ok)
    test.validate()
